"""La Marzocco Cloud API Client."""

from __future__ import annotations

import asyncio
import logging
import time
from datetime import datetime
from http import HTTPMethod
from typing import Any

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from .const import (
    CUSTOMER_URL,
    DEFAULT_CLIENT_ID,
    DEFAULT_CLIENT_SECRET,
    GW_AWS_PROXY_BASE_URL,
    GW_MACHINE_BASE_URL,
    LOGOUT_URL,
    TOKEN_URL,
    BoilerType,
    FirmwareType,
    PhysicalKey,
    PrebrewMode,
    SmartStandbyMode,
)
from .helpers import is_success
from .exceptions import AuthFail, RequestNotSuccessful
from .models import (
    AccessToken,
    LaMarzoccoDeviceInfo,
    LaMarzoccoFirmware,
    LaMarzoccoWakeUpSleepEntry,
)

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoCloudClient:
    """La Marzocco Cloud Client."""

    _client: ClientSession

    def __init__(
        self, username: str, password: str, client: ClientSession | None = None
    ) -> None:
        if client is None:
            self._client = ClientSession()
        else:
            self._client = client
        self._username = username
        self._password = password
        self._access_token: AccessToken | None = None

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        if self._access_token is None or self._access_token.expires_in < time.time():
            return await self._async_get_access_token()
        if self._access_token.expires_in < time.time() + 300:
            return await self._async_get_refresh_token()
        return self._access_token.access_token

    async def _async_get_access_token(self) -> str:
        """Get a new access token."""
        data = {
            "username": self._username,
            "password": self._password,
            "grant_type": "password",
            "client_id": DEFAULT_CLIENT_ID,
            "client_secret": DEFAULT_CLIENT_SECRET,
        }
        _LOGGER.debug("Getting new access token, data: %s", data)
        return await self.__async_get_token(data)

    async def _async_get_refresh_token(self) -> str:
        """Refresh the access token."""
        assert self._access_token is not None
        data = {
            "refresh_token": self._access_token.refresh_token,
            "grant_type": "refresh_token",
            "client_id": DEFAULT_CLIENT_ID,
            "client_secret": DEFAULT_CLIENT_SECRET,
        }
        _LOGGER.debug("Refreshing access token, data: %s", data)
        return await self.__async_get_token(data)

    async def __async_get_token(self, data: dict[str, Any]) -> str:
        """Wrapper for a token request."""
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        try:
            response = await self._client.post(TOKEN_URL, data=data, headers=headers)
        except ClientError as ex:
            raise RequestNotSuccessful(
                "Error during HTTP request."
                + f"Request to endpoint {TOKEN_URL} failed with error: {ex}"
            ) from ex
        if is_success(response):
            json_response = await response.json()
            self._access_token = AccessToken(
                access_token=json_response["access_token"],
                refresh_token=json_response["refresh_token"],
                expires_in=time.time() + json_response["expires_in"],
            )
            _LOGGER.debug("Got new access token: %s", json_response)
            return json_response["access_token"]

        if response.status == 401:
            raise AuthFail("Invalid username or password")

        raise RequestNotSuccessful(
            f"Request to endpoint {TOKEN_URL} failed with status code {response.status}"
            + f"response: {await response.text()}"
        )

    async def async_logout(self) -> None:
        """Logout from the cloud."""
        if self._access_token is None:
            return
        try:
            response = await self._client.post(LOGOUT_URL, data={})
        except ClientError as ex:
            raise RequestNotSuccessful(
                "Error during HTTP request."
                + f"Request to endpoint {LOGOUT_URL} failed with error: {ex}"
            ) from ex
        if not is_success(response):
            raise RequestNotSuccessful(
                f"Request to endpoint {LOGOUT_URL} failed with status code {response.status},"
                + "response: {await response.text()}"
            )
        self._access_token = None

    async def _rest_api_call(
        self,
        url: str,
        method: HTTPMethod,
        data: dict[str, Any] | None = None,
        timeout: int = 5,
    ) -> Any:
        """Wrapper for the API call."""

        access_token = await self.async_get_access_token()
        headers = {"Authorization": f"Bearer {access_token}"}

        try:
            response = await self._client.request(
                method=method,
                url=url,
                json=data,
                timeout=ClientTimeout(total=timeout),
                headers=headers,
            )
        except ClientError as ex:
            raise RequestNotSuccessful(
                f"Error during HTTP request. Request to endpoint {url} failed with error: {ex}"
            ) from ex

        # ensure status code indicates success
        if is_success(response):
            json_response = await response.json()
            _LOGGER.debug("Request to %s successful", json_response)
            return json_response["data"]

        raise RequestNotSuccessful(
            f"Request to endpoint {response.url} failed with status code {response.status}"
            + f"response: {await response.text()}"
        )

    async def get_customer_fleet(self) -> dict[str, LaMarzoccoDeviceInfo]:
        """Get basic machine info from the customer endpoint."""

        machine_info: dict[str, LaMarzoccoDeviceInfo] = {}

        data = await self._rest_api_call(url=CUSTOMER_URL, method=HTTPMethod.GET)
        fleet = data.get("fleet", [])
        for machine_data in fleet:
            key = machine_data.get("communicationKey")
            name = machine_data.get("name")

            machine = machine_data.get("machine", {})
            serial_number = machine.get("serialNumber")
            model_name = machine.get("model", {}).get("name")

            machine_info[serial_number] = LaMarzoccoDeviceInfo(
                serial_number=serial_number,
                name=name,
                communication_key=key,
                model=model_name,
            )

        return machine_info

    async def get_config(self, serial_number: str) -> dict[str, Any]:
        """Get configuration from cloud"""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/configuration"
        return await self._rest_api_call(url=url, method=HTTPMethod.GET)

    async def set_power(
        self,
        serial_number: str,
        enabled: bool,
    ) -> bool:
        """Turn power of machine on or off"""

        mode = "BrewingMode" if enabled else "StandBy"

        data = {"status": mode}
        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/status"
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def set_steam(
        self,
        serial_number: str,
        enabled: bool,
    ) -> bool:
        """Turn Steamboiler on or off"""

        data = {
            "identifier": BoilerType.STEAM.value,
            "state": enabled,
        }
        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/enable-boiler"
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def set_temp(
        self,
        serial_number: str,
        boiler: BoilerType,
        temperature: float,
    ) -> bool:
        """Set boiler temperature (in Celsius)."""

        data = {"identifier": boiler.value, "value": temperature}
        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/target-boiler"
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True

        return False

    async def set_prebrew_mode(
        self,
        serial_number: str,
        mode: PrebrewMode,
    ) -> bool:
        """Enable/Disable Pre-Brew or Pre-Infusion (mutually exclusive)."""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/enable-preinfusion"
        data = {"mode": mode.value}
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def configure_pre_brew_infusion_time(
        self,
        serial_number: str,
        on_time: float,
        off_time: float,
        key: PhysicalKey,
    ) -> bool:
        """Set Pre-Brew details. Also used for preinfusion (prebrewOnTime=0, prebrewOnTime=ms)."""

        on_time = round(on_time, 1) * 1000
        off_time = round(off_time, 1) * 1000
        button = f"Dose{key.name}"

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/setting-preinfusion"
        data = {
            "button": button,
            "group": "Group1",
            "holdTimeMs": int(off_time),
            "wetTimeMs": int(on_time),
        }
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def enable_plumbin(
        self,
        serial_number: str,
        enable: bool,
    ) -> bool:
        """Enable or disable plumbin mode"""

        data = {"enable": enable}
        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/enable-plumbin"
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def set_dose(
        self,
        serial_number: str,
        key: PhysicalKey,
        value: int,
    ) -> bool:
        """Set the value for a dose"""

        dose_index = f"Dose{key.name}"

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/dose"
        data = {
            "dose_index": dose_index,
            "dose_type": "PulsesType",
            "group": "Group1",
            "value": value,
        }

        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def set_dose_hot_water(
        self,
        serial_number: str,
        value: int,
    ) -> bool:
        """Set the value for the hot water dose"""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/dose-tea"
        data = {"dose_index": "DoseA", "value": value}
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    # async def set_schedule(
    #     self,
    #     serial_number: str,
    #     schedule: LaMarzoccoCloudSchedule,
    # ) -> bool:
    #     """Set auto-on/off schedule"""

    #     url = f"{GW_MACHINE_BASE_URL}/{serial_number}/scheduling"
    #     response = await self._rest_api_call(
    #         url=url, method=HTTPMethod.POST, data=dict(schedule)
    #     )
    #     if await self._check_cloud_command_status(serial_number, response):
    #         return True
    #     return False

    async def set_wake_up_sleep(
        self, serial_number: str, wake_up_sleep_entry: LaMarzoccoWakeUpSleepEntry
    ) -> bool:
        """Enable or disable wake-up sleep mode"""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/wake-up-sleep/{wake_up_sleep_entry.entry_id}"
        data = {
            "days": [day.value for day in wake_up_sleep_entry.days],
            "enable": wake_up_sleep_entry.enabled,
            "id": wake_up_sleep_entry.entry_id,
            "steam": wake_up_sleep_entry.steam,
            "timeOff": wake_up_sleep_entry.time_off,
            "timeOn": wake_up_sleep_entry.time_on,
        }
        response = await self._rest_api_call(url=url, method=HTTPMethod.PUT, data=data)
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def set_smart_standby(
        self,
        serial_number: str,
        enabled: bool,
        minutes: int,
        mode: SmartStandbyMode,
    ) -> bool:
        """Set the smart standby configuration"""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/smart-standby"
        response = await self._rest_api_call(
            url=url,
            method=HTTPMethod.POST,
            data={
                "enabled": enabled,
                "minutes": minutes,
                "mode": mode.value,
            },
        )
        if await self._check_cloud_command_status(serial_number, response):
            return True
        return False

    async def start_backflush(self, serial_number: str) -> None:
        """Send command to start backflushing"""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/enable-backflush"
        data = {"enable": True}
        await self._rest_api_call(url=url, method=HTTPMethod.POST, data=data)

    async def token_command(self, serial_number: str) -> None:
        """Send token request command to cloud. This is needed when the local API returns 403."""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/token-request"
        response = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        await self._check_cloud_command_status(serial_number, response)

    async def _check_cloud_command_status(
        self,
        serial_number: str,
        command_response: dict[str, Any],
    ) -> bool:
        """Check the status of a cloud command"""

        if command_id := command_response.get("commandId"):
            _LOGGER.debug("Checking status of command %s", command_id)
            url = f"{GW_AWS_PROXY_BASE_URL}/{serial_number}/commands/{command_id}"
            counter = 0
            status = "PENDING"
            while status == "PENDING" and counter < 5:
                await asyncio.sleep(1)  # give a second to settle in
                response = await self._rest_api_call(url=url, method=HTTPMethod.GET)
                if response is None:
                    return False
                status = response.get("status", "PENDING")
                if status == "PENDING":
                    _LOGGER.debug("Command %s still pending", command_id)
                    counter += 1
                    continue
                if status == "COMPLETED":
                    response_payload = response.get("responsePayload")
                    if response_payload is None:
                        return False
                    _LOGGER.debug("Command %s completed", command_id)
                    return response_payload.get("status") == "success"
        _LOGGER.debug("Command %s failed", command_id)
        return False

    async def get_firmware(
        self,
        serial_number: str,
    ) -> dict[FirmwareType, LaMarzoccoFirmware]:
        """Get Firmware details."""

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/firmware/"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        firmware: dict[FirmwareType, LaMarzoccoFirmware] = {}
        for component in FirmwareType:
            current_version = result.get(f"{component}_firmware", {}).get("version")
            latest_version = result.get(f"{component}_firmware", {}).get(
                "targetVersion"
            )
            firmware[component] = LaMarzoccoFirmware(
                current_version=current_version,
                latest_version=latest_version,
            )
        return firmware

    async def update_firmware(
        self, serial_number: str, component: FirmwareType
    ) -> bool:
        """Update Firmware."""

        _LOGGER.debug("Updating firmware for component %s", component)
        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/firmware/{component.value}/update"
        await self._rest_api_call(url=url, method=HTTPMethod.POST, data={})
        retry_counter = 0
        while retry_counter <= 20:
            firmware = await self.get_firmware(serial_number)

            if (
                firmware[component].current_version
                == firmware[component].latest_version
            ):
                _LOGGER.debug("Firmware update for component %s successful", component)
                return True
            _LOGGER.debug(
                "Firmware update for component %s still in progress", component
            )
            await asyncio.sleep(15)
            retry_counter += 1
        _LOGGER.debug(
            "Firmware update for component %s timed out waiting to finish",
            component,
        )
        return False

    async def get_statistics(self, serial_number: str) -> list[dict[str, Any]]:
        """Get statistics from cloud."""

        _LOGGER.debug("Getting statistics from cloud")

        url = f"{GW_MACHINE_BASE_URL}/{serial_number}/statistics/counters"

        return await self._rest_api_call(url=url, method=HTTPMethod.GET)

    async def get_daily_statistics(
        self,
        serial_number: str,
        start_date: datetime,
        end_date: datetime,
        timezone_offset: int,
        timezone: str,
    ) -> list[dict[str, Any]]:
        """Get daily statistics from cloud."""

        _LOGGER.debug("Getting daily statistics from cloud")

        url = (
            f"{GW_MACHINE_BASE_URL}/{serial_number}/statistics/daily"
            + f"?startDate={start_date.isoformat()}"
            + f"&endDate={end_date.isoformat()}"
            + f"&timezoneOffset={timezone_offset}"
            + f"&timezone={timezone}"
        )

        return await self._rest_api_call(url=url, method=HTTPMethod.GET, timeout=60)



"""Import for ease of use."""

from .client_bluetooth import LaMarzoccoBluetoothClient
from .client_cloud import LaMarzoccoCloudClient
from .client_local import LaMarzoccoLocalClient
from .lm_grinder import LaMarzoccoGrinder
from .lm_machine import LaMarzoccoMachine



""" Helper functions for pylamarzocco. """

from typing import Any
from aiohttp import ClientResponse

from .const import (
    BoilerType,
    FirmwareType,
    PhysicalKey,
    PrebrewMode,
    SmartStandbyMode,
    WeekDay,
)
from .models import (
    LaMarzoccoBoiler,
    LaMarzoccoCoffeeStatistics,
    LaMarzoccoFirmware,
    LaMarzoccoPrebrewConfiguration,
    LaMarzoccoSmartStandby,
    LaMarzoccoWakeUpSleepEntry,
)

# def schedule_to_request(schedule: LaMarzoccoSchedule) -> LaMarzoccoCloudSchedule:
#     """convert schedule to API expected input format"""

#     schedule_conv: LaMarzoccoCloudSchedule = {
#         "enable": schedule.enabled,
#         "days": [],
#     }
#     for day, schedule_day in schedule.days.items():
#         # requests wants 00:00, response gives 24:00
#         h_on = "00" if schedule_day.h_on == 24 else str(schedule_day.h_on).zfill(2)
#         h_off = "00" if schedule_day.h_off == 24 else str(schedule_day.h_off).zfill(2)

#         hh_mm_on = h_on + ":" + str(schedule_day.m_on).zfill(2)
#         hh_mm_off = h_off + ":" + str(schedule_day.m_off).zfill(2)

#         schedule_conv["days"].append(
#             {
#                 "day": str.upper(day),
#                 "enable": schedule_day.enabled,
#                 "on": hh_mm_on,
#                 "off": hh_mm_off,
#             }
#         )
#     return schedule_conv


# def parse_schedule(schedule: dict[str, Any]) -> LaMarzoccoSchedule:
#     """Parse schedule from API config object."""

#     global_enable: bool = schedule["enabled"]
#     days: dict[WeekDay, LaMarzoccoScheduleDay] = {}
#     for weekday in WeekDay:
#         day_settings = schedule[weekday]
#         days[weekday] = LaMarzoccoScheduleDay(
#             enabled=day_settings["enabled"],
#             h_on=day_settings["h_on"],
#             h_off=day_settings["h_off"],
#             m_on=day_settings["m_on"],
#             m_off=day_settings["m_off"],
#         )
#     return LaMarzoccoSchedule(enabled=global_enable, days=days)


def parse_boilers(boilers: list[dict[str, Any]]) -> dict[BoilerType, LaMarzoccoBoiler]:
    """Parse boiler settings from API config object."""

    parsed_boilers: dict[BoilerType, LaMarzoccoBoiler] = {}
    for boiler in boilers:
        is_enabled = boiler["isEnabled"]
        current_temp = boiler["current"]
        target_temp = boiler["target"]
        parsed_boilers[BoilerType(boiler["id"])] = LaMarzoccoBoiler(
            enabled=is_enabled,
            current_temperature=current_temp,
            target_temperature=target_temp,
        )
    return parsed_boilers


def parse_preinfusion_settings(
    config: dict[str, Any]
) -> tuple[PrebrewMode, dict[PhysicalKey, LaMarzoccoPrebrewConfiguration]]:
    """Parse preinfusion settings from API config object."""

    parsed: dict[PhysicalKey, LaMarzoccoPrebrewConfiguration] = {}
    i = 1
    preinfusion_settings = config.get("preinfusionSettings", {})
    mode = PrebrewMode(preinfusion_settings.get("mode", "Disabled"))
    for group in preinfusion_settings.get("Group1", {}):
        parsed[PhysicalKey(i)] = LaMarzoccoPrebrewConfiguration(
            on_time=group.get("preWetTime", 0),
            off_time=group.get("preWetHoldTime", 0),
        )
        i += 1

    return mode, parsed


def parse_coffee_doses(config: dict[str, Any]) -> tuple[dict[PhysicalKey, float], int]:
    """Parse doses from API config object."""

    parsed: dict[PhysicalKey, float] = {}
    i = 1
    group_capabilities = config.get("groupCapabilities", [])
    if len(group_capabilities) == 0:
        return parsed, 0

    for dose in group_capabilities[0].get("doses", []):
        parsed[PhysicalKey(i)] = dose.get("stopTarget", 0)
        i += 1
    dose_hot_water: int = (
        config.get("teaDoses", {}).get("DoseA", {}).get("stopTarget", 0)
    )
    return parsed, dose_hot_water


def parse_cloud_statistics(
    statistics: list[dict[str, Any]]
) -> LaMarzoccoCoffeeStatistics:
    """Parse statistics from API statistics object."""

    drink_stats: dict[PhysicalKey, int] = {}
    continuous = 0
    total_flushes = 0

    for stat in statistics:
        coffee_type: int = stat["coffeeType"]
        count: int = stat["count"]
        if 0 <= coffee_type < 4:
            drink_stats[PhysicalKey(coffee_type + 1)] = count
        elif coffee_type == 4:
            continuous = count
        elif coffee_type == -1:
            total_flushes = count
    return LaMarzoccoCoffeeStatistics(
        drink_stats=drink_stats,
        continous=continuous,
        total_flushes=total_flushes,
    )


def parse_firmware(
    raw_firmware: list[dict[str, Any]],
    current_firmware: dict[FirmwareType, LaMarzoccoFirmware] | None = None,
) -> dict[FirmwareType, LaMarzoccoFirmware]:
    """Parse firmware from API config object."""

    parsed = {}
    for fw in raw_firmware:
        fw_type = FirmwareType(fw["name"].split("_")[0])
        version = fw["fw_version"]
        latest_version = (
            version
            if not current_firmware
            else current_firmware[fw_type].latest_version
        )
        parsed[fw_type] = LaMarzoccoFirmware(
            current_version=version,
            latest_version=latest_version,
        )
    return parsed


def parse_webhook_statistics(statistics: dict[str, Any]) -> LaMarzoccoCoffeeStatistics:
    """Parse statistics from webhook statistics object."""

    continuous = 0
    group = statistics["groups"][0]
    doses = group["doses"]
    drink_stats: dict[PhysicalKey, int] = {}
    for dose in doses:
        key: str = next(iter(dose))
        if key.startswith("Dose"):
            drink_stats[PhysicalKey(ord(key[-1]) - 64)] = dose[key]
        elif key == "ContinuousDose":
            continuous = dose[key]
    total_flushing = group["clean"]
    return LaMarzoccoCoffeeStatistics(
        drink_stats=drink_stats,
        continous=continuous,
        total_flushes=total_flushing,
    )


def parse_smart_standby(smart_standby_config: dict[str, Any]) -> LaMarzoccoSmartStandby:
    """Parse smart standby mode from API config object."""
    return LaMarzoccoSmartStandby(
        enabled=smart_standby_config.get("enabled", False),
        minutes=smart_standby_config.get("minutes", 10),
        mode=SmartStandbyMode(smart_standby_config.get("mode", "PowerOn")),
    )


def parse_wakeup_sleep_entries(
    entries: list[dict[str, Any]]
) -> dict[str, LaMarzoccoWakeUpSleepEntry]:
    """Parse wake up sleep entries from API config object."""
    parsed = {}
    for entry in entries:
        wake_up_sleep_entry = LaMarzoccoWakeUpSleepEntry(
            enabled=entry["enabled"],
            days=[WeekDay(day) for day in entry["days"]],
            entry_id=entry["id"],
            steam=entry["steam"],
            time_on=entry["timeOn"],
            time_off=entry["timeOff"],
        )
        parsed[wake_up_sleep_entry.entry_id] = wake_up_sleep_entry
    return parsed


def is_success(response: ClientResponse) -> bool:
    """Check if response is successful."""
    return 200 <= response.status < 300



"""Bluetooth class for La Marzocco machines."""

from __future__ import annotations

import base64
import json
import logging
from typing import Any

from bleak import (
    BaseBleakScanner,
    BleakClient,
    BleakError,
    BleakScanner,
    BLEDevice,
)

from .const import (
    AUTH_CHARACTERISTIC,
    BT_MODEL_PREFIXES,
    SETTINGS_CHARACTERISTIC,
    BoilerType,
)
from .exceptions import BluetoothConnectionFailed

_logger = logging.getLogger(__name__)


class LaMarzoccoBluetoothClient:
    """Class to interact with machine via Bluetooth."""

    def __init__(
        self,
        username: str,
        serial_number: str,
        token: str,
        address_or_ble_device: BLEDevice | str,
    ) -> None:
        """Initializes a new LaMarzoccoBluetoothClient instance."""
        self._username = username
        self._serial_number = serial_number
        self._token = token
        self._address = (
            address_or_ble_device.address
            if isinstance(address_or_ble_device, BLEDevice)
            else address_or_ble_device
        )
        self._address_or_ble_device = address_or_ble_device
        self._client = BleakClient(address_or_ble_device)

    @staticmethod
    async def discover_devices(
        scanner: BaseBleakScanner | BleakScanner | None = None,
    ) -> list[BLEDevice]:
        """Find machines based on model name."""
        ble_devices: list[BLEDevice] = []

        if scanner is None:
            scanner = BleakScanner()
        assert hasattr(scanner, "discover")
        devices: list[BLEDevice] = await scanner.discover()
        for device in devices:
            if device.name and device.name.startswith(BT_MODEL_PREFIXES):
                ble_devices.append(device)

        return ble_devices

    @property
    def address(self) -> str:
        """Return the BT MAC address of the machine."""

        return self._address

    @property
    def connected(self) -> bool:
        """Return the connection status."""

        return self._client.is_connected

    async def set_power(self, enabled: bool) -> None:
        """Power on the machine."""

        mode = "BrewingMode" if enabled else "StandBy"
        data = {
            "name": "MachineChangeMode",
            "parameter": {
                "mode": mode,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def set_steam(self, enabled: bool) -> None:
        """Power cycle steam."""

        data = {
            "name": "SettingBoilerEnable",
            "parameter": {
                "identifier": "SteamBoiler",
                "state": enabled,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def set_temp(self, boiler: BoilerType, temperature: float) -> None:
        """Set boiler temperature (in Celsius)"""

        data = {
            "name": "SettingBoilerTarget",
            "parameter": {
                "identifier": boiler.value,
                "value": temperature,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def _write_bluetooth_message(
        self, characteristic: str, message: bytes | str
    ) -> None:
        """Connect to machine and write a message."""

        if not self._client.is_connected:
            try:
                self._client = BleakClient(self._address_or_ble_device)
                await self._client.connect()
                await self._authenticate()
            except (BleakError, TimeoutError) as e:
                raise BluetoothConnectionFailed(
                    f"Failed to connect to machine with Bluetooth: {e}"
                ) from e

        # check if message is already bytes string
        if not isinstance(message, bytes):
            message = bytes(message, "utf-8")

        # append trailing zeros to settings message
        if characteristic == SETTINGS_CHARACTERISTIC:
            message += b"\x00"

        _logger.debug("Sending bluetooth message: %s to %s", message, characteristic)

        await self._client.write_gatt_char(characteristic, message)

    async def _write_bluetooth_json_message(
        self,
        data: dict[str, Any],
        characteristic: str = SETTINGS_CHARACTERISTIC,
    ) -> None:
        """Write a json message to the machine."""

        await self._write_bluetooth_message(
            characteristic=characteristic,
            message=json.dumps(data, separators=(",", ":")),
        )

    async def _authenticate(self) -> None:
        """Build authentication string and send it to the machine."""

        user = self._username + ":" + self._serial_number
        user_bytes = user.encode("utf-8")
        token = self._token.encode("utf-8")
        auth_string = base64.b64encode(user_bytes) + b"@" + base64.b64encode(token)
        await self._write_bluetooth_message(
            characteristic=AUTH_CHARACTERISTIC,
            message=auth_string,
        )



"""Module for La Marzocco coffee machine."""

from __future__ import annotations

import json
import logging
from collections.abc import Callable
from datetime import datetime
from typing import Any

from .client_bluetooth import LaMarzoccoBluetoothClient
from .client_cloud import LaMarzoccoCloudClient
from .client_local import LaMarzoccoLocalClient
from .const import (
    BoilerType,
    FirmwareType,
    MachineModel,
    PhysicalKey,
    PrebrewMode,
    SmartStandbyMode,
    SteamLevel,
)
from .exceptions import ClientNotInitialized, UnknownWebSocketMessage
from .helpers import (
    parse_boilers,
    parse_cloud_statistics,
    parse_coffee_doses,
    parse_preinfusion_settings,
    parse_smart_standby,
    parse_wakeup_sleep_entries,
)
from .lm_device import LaMarzoccoDevice
from .models import (
    LaMarzoccoCoffeeStatistics,
    LaMarzoccoMachineConfig,
    LaMarzoccoSmartStandby,
    LaMarzoccoWakeUpSleepEntry,
)

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoMachine(LaMarzoccoDevice):
    """Class for La Marzocco coffee machine"""

    def __init__(
        self,
        model: MachineModel,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
        bluetooth_client: LaMarzoccoBluetoothClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoCoffeeMachine instance"""

        super().__init__(
            model=model,
            serial_number=serial_number,
            name=name,
            cloud_client=cloud_client,
            local_client=local_client,
            bluetooth_client=bluetooth_client,
        )
        self.config: LaMarzoccoMachineConfig = LaMarzoccoMachineConfig(
            turned_on=False,
            boilers={},
            prebrew_mode=PrebrewMode.DISABLED,
            plumbed_in=False,
            prebrew_configuration={},
            dose_hot_water=0,
            doses={},
            water_contact=False,
            brew_active=False,
            backflush_enabled=False,
            brew_active_duration=0,
            smart_standby=LaMarzoccoSmartStandby(
                enabled=False,
                minutes=10,
                mode=SmartStandbyMode.POWER_ON,
            ),
            wake_up_sleep_entries={},
        )
        self.statistics: LaMarzoccoCoffeeStatistics = LaMarzoccoCoffeeStatistics(
            drink_stats={},
            continous=0,
            total_flushes=0,
        )

        self._notify_callback: Callable[[], None] | None = None
        self._system_info: dict[str, Any] | None = None
        self._timestamp_last_websocket_msg: datetime | None = None

    @classmethod
    async def create(cls, *args: Any, **kwargs: Any) -> LaMarzoccoMachine:
        """Create a new LaMarzoccoMachine instance"""

        self = cls(*args, **kwargs)
        await self.get_config()
        if self._cloud_client is not None:
            await self.get_firmware()
            await self.get_statistics()
        return self

    @property
    def full_model_name(self) -> str:
        """Return the full model name"""

        if self.model == MachineModel.LINEA_MICRA:
            return "Linea Micra"

        return self.model

    @property
    def websocket_connected(self) -> bool:
        """Return the connection status of the websocket client."""
        if self._local_client and self._local_client.websocket:
            return not self._local_client.websocket.closed
        return False

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        super().parse_config(raw_config)
        self._raw_config = raw_config
        self.config.turned_on = raw_config["machineMode"] == "BrewingMode"
        self.config.plumbed_in = raw_config["isPlumbedIn"]
        self.config.water_contact = raw_config["tankStatus"]
        self.config.backflush_enabled = raw_config["isBackFlushEnabled"]
        self.config.doses, self.config.dose_hot_water = parse_coffee_doses(raw_config)
        self.config.boilers = parse_boilers(raw_config["boilers"])
        self.config.prebrew_mode, self.config.prebrew_configuration = (
            parse_preinfusion_settings(raw_config)
        )
        self.config.smart_standby = parse_smart_standby(
            raw_config.get("smartStandBy", {})
        )
        self.config.wake_up_sleep_entries = parse_wakeup_sleep_entries(
            raw_config.get("wakeUpSleepEntries", {})
        )

    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

        self.statistics = parse_cloud_statistics(raw_statistics)

    async def set_power(
        self,
        enabled: bool,
    ) -> bool:
        """Turn power of machine on or off"""

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_power",
            enabled=enabled,
        ):
            self.config.turned_on = enabled
            self.config.boilers[BoilerType.COFFEE].enabled = enabled
            return True
        return False

    async def set_steam(
        self,
        enabled: bool,
    ) -> bool:
        """Turn Steamboiler on or off"""

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_steam",
            enabled=enabled,
        ):
            self.config.boilers[BoilerType.STEAM].enabled = enabled
            return True
        return False

    async def set_temp(
        self,
        boiler: BoilerType,
        temperature: float,
    ) -> bool:
        """Set target temperature for boiler"""

        if boiler == BoilerType.STEAM:
            if self.model == MachineModel.LINEA_MICRA:
                if temperature not in SteamLevel:
                    msg = "Steam temp must be one of 126, 128, 131 (°C)"
                    _LOGGER.debug(msg)
                    raise ValueError(msg)
            elif self.model == MachineModel.LINEA_MINI:
                _LOGGER.error("Steam temp is not supported on Linea Mini.")
                return False
        else:
            if temperature > 104 or temperature < 85:
                msg = "Coffee temp must be between 85 and 104 (°C)"
                _LOGGER.debug(msg)
                raise ValueError(msg)

            temperature = round(temperature, 1)

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_temp",
            boiler=boiler,
            temperature=temperature,
        ):

            self.config.boilers[boiler].target_temperature = temperature
            return True
        return False

    async def set_steam_level(
        self,
        level: SteamLevel,
    ) -> bool:
        """Set steam level"""
        return await self.set_temp(BoilerType.STEAM, level)

    async def set_prebrew_mode(self, mode: PrebrewMode) -> bool:
        """Set preinfusion mode"""

        if mode == PrebrewMode.PREINFUSION and not self.config.plumbed_in:
            msg = "Pre-Infusion can only be enabled when plumbin is enabled."
            _LOGGER.debug(msg)
            raise ValueError(msg)

        if await self.cloud_client.set_prebrew_mode(self.serial_number, mode):
            self.config.prebrew_mode = mode
            return True
        return False

    async def set_prebrew_time(
        self,
        prebrew_on_time: float | None = None,
        prebrew_off_time: float | None = None,
        key: PhysicalKey = PhysicalKey.A,
    ) -> bool:
        """Set prebrew time"""

        if prebrew_on_time is None:
            prebrew_on_time = self.config.prebrew_configuration[key].on_time

        if prebrew_off_time is None:
            prebrew_off_time = self.config.prebrew_configuration[key].off_time

        if await self.cloud_client.configure_pre_brew_infusion_time(
            self.serial_number, prebrew_on_time, prebrew_off_time, key
        ):
            self.config.prebrew_configuration[key].on_time = prebrew_on_time
            self.config.prebrew_configuration[key].off_time = prebrew_off_time
            return True
        return False

    async def set_preinfusion_time(
        self,
        preinfusion_time: float,
        key: PhysicalKey = PhysicalKey.A,
    ) -> bool:
        """Set preinfusion time"""

        if await self.cloud_client.configure_pre_brew_infusion_time(
            self.serial_number, 0, preinfusion_time, key
        ):
            self.config.prebrew_configuration[key].off_time = preinfusion_time
            return True
        return False

    async def set_dose(self, dose: int, key: PhysicalKey) -> bool:
        """Set dose"""

        if await self.cloud_client.set_dose(self.serial_number, key, dose):
            self.config.doses[key] = dose
            return True
        return False

    async def set_dose_tea_water(self, dose: int) -> bool:
        """Set tea dose"""

        if await self.cloud_client.set_dose_hot_water(self.serial_number, dose):
            self.config.dose_hot_water = dose
            return True
        return False

    async def set_plumbed_in(self, enabled: bool) -> bool:
        """Set plumbed in"""

        if await self.cloud_client.enable_plumbin(self.serial_number, enabled):
            self.config.plumbed_in = enabled
            return True
        return False

    async def start_backflush(self) -> None:
        """Start backflush"""

        await self.cloud_client.start_backflush(self.serial_number)
        self.config.backflush_enabled = True

    # async def set_schedule(self, schedule: LaMarzoccoSchedule) -> bool:
    #     """Set schedule"""

    #     if await self.cloud_client.set_schedule(
    #         self.serial_number, schedule_to_request(schedule)
    #     ):
    #         self.config.auto_on_off_schedule = schedule
    #         return True
    #     return False

    # async def enable_schedule_globally(self, enabled: bool) -> bool:
    #     """Enable schedule globally"""

    #     schedule = deepcopy(self.config.auto_on_off_schedule)
    #     schedule.enabled = enabled
    #     return await self.set_schedule(schedule)

    # async def set_schedule_day(
    #     self,
    #     day: WeekDay,
    #     enabled: bool,
    #     h_on: int,
    #     m_on: int,
    #     h_off: int,
    #     m_off: int,
    # ) -> bool:
    #     """Configure a single day in the schedule"""

    #     day_settings = LaMarzoccoScheduleDay(
    #         enabled=enabled,
    #         h_on=h_on,
    #         h_off=h_off,
    #         m_on=m_on,
    #         m_off=m_off,
    #     )
    #     schedule = deepcopy(self.config.auto_on_off_schedule)
    #     schedule.days[day] = day_settings
    #     return await self.set_schedule(schedule)

    async def set_wake_up_sleep(self, wake_up_sleep_entry: LaMarzoccoWakeUpSleepEntry):
        """Set wake up sleep"""

        if await self.cloud_client.set_wake_up_sleep(
            self.serial_number, wake_up_sleep_entry
        ):
            self.config.wake_up_sleep_entries[wake_up_sleep_entry.entry_id] = (
                wake_up_sleep_entry
            )
            return True
        return False

    async def set_smart_standby(
        self, enabled: bool, minutes: int, mode: SmartStandbyMode
    ) -> bool:
        """Set smart standby"""

        if await self.cloud_client.set_smart_standby(
            serial_number=self.serial_number,
            enabled=enabled,
            minutes=minutes,
            mode=mode,
        ):
            self.config.smart_standby = LaMarzoccoSmartStandby(
                enabled=enabled, mode=mode, minutes=minutes
            )
            return True
        return False

    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""

        return await self.cloud_client.update_firmware(self.serial_number, component)

    async def websocket_connect(
        self, notify_callback: Callable[[], None] | None = None
    ) -> None:
        """Connect to the websocket of the machine."""

        self._notify_callback = notify_callback
        if self._local_client is None:
            raise ClientNotInitialized("Local client not initialized")

        await self._local_client.websocket_connect(
            callback=self.on_websocket_message_received
        )

    async def websocket_disconnect(self) -> None:
        """Disconnect from the websocket of the machine."""

        if self._local_client is None or self._local_client.websocket is None:
            return

        await self._local_client.websocket.close()

    def on_websocket_message_received(self, message: str | bytes) -> None:
        """Websocket message received"""

        self._timestamp_last_websocket_msg = datetime.now()
        message = str(message)

        _LOGGER.debug("Received message from websocket, message %s", message)
        notify = False
        try:
            notify = self._parse_websocket_message(message)
        except UnknownWebSocketMessage as exc:
            _LOGGER.warning("Unknown websocket message received")
            _LOGGER.warning("Message: %s", message)
            _LOGGER.debug(exc)
        except TypeError as exc:
            _LOGGER.error("Error when parsing websocket message: %s, %s", message, exc)

        if notify and self._notify_callback:
            self._notify_callback()

    def _parse_websocket_message(self, message: str) -> bool:
        """Handle a message received on the websocket."""

        message = json.loads(message)

        if isinstance(message, dict):
            self._parse_dict_message(message)
            return False

        if isinstance(message, list):
            if "KeepAlive" in message[0]:
                return False
            self._parse_list_message(message)
            return True

        raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")

    def _parse_dict_message(self, message: Any) -> bool:
        """ "Parse websocket message that is a dict."""

        if "MachineConfiguration" in message:
            # got machine configuration
            self._raw_config = json.loads(message["MachineConfiguration"])
            return False

        if "SystemInfo" in message:
            self._system_info = json.loads(message["SystemInfo"])
            return False

        raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")

    def _parse_list_message(self, message: list[dict[str, Any]]) -> bool:
        """Parse websocket message that is a list."""

        property_updated = False
        for msg in message:

            if "SteamBoilerUpdateTemperature" in msg:
                self.config.boilers[BoilerType.STEAM].current_temperature = msg[
                    "SteamBoilerUpdateTemperature"
                ]
                property_updated = True

            elif "CoffeeBoiler1UpdateTemperature" in msg:
                self.config.boilers[BoilerType.COFFEE].current_temperature = msg[
                    "CoffeeBoiler1UpdateTemperature"
                ]
                property_updated = True

            elif "Sleep" in msg:
                self.config.turned_on = False
                property_updated = True

            elif "SteamBoilerEnabled" in msg:
                value = msg["SteamBoilerEnabled"]
                self.config.boilers[BoilerType.STEAM].enabled = value
                property_updated = True

            elif "WakeUp" in msg:
                self.config.turned_on = True
                property_updated = True

            elif "MachineMode" in msg:
                self.config.turned_on = msg["MachineMode"] == "BrewingMode"
                property_updated = True

            elif "MachineStatistics" in msg:
                property_updated = True

            elif "BrewingUpdateGroup1Time" in msg:
                self.config.brew_active_duration = msg["BrewingUpdateGroup1Time"]
                property_updated = True

            elif "BrewingStartedGroup1StopType" in msg:
                self.config.brew_active = True
                property_updated = True

            elif (
                "BrewingStoppedGroup1StopType" in msg or "BrewingSnapshotGroup1" in msg
            ):
                self.config.brew_active = False
                property_updated = True

            elif "SteamBoilerUpdateSetPoint" in msg:
                self.config.boilers[BoilerType.STEAM].target_temperature = msg[
                    "SteamBoilerUpdateSetPoint"
                ]
                property_updated = True

            elif "CoffeeBoiler1UpdateSetPoint" in msg:
                self.config.boilers[BoilerType.COFFEE].target_temperature = msg[
                    "CoffeeBoiler1UpdateSetPoint"
                ]
                property_updated = True

            elif "BoilersTargetTemperature" in msg:
                target_temps = json.loads(msg["BoilersTargetTemperature"])
                for boiler in BoilerType:
                    self.config.boilers[boiler].target_temperature = target_temps[
                        boiler
                    ]
                property_updated = True

            elif "Boilers" in msg:
                boilers = json.loads(msg["Boilers"])
                self.config.boilers = parse_boilers(boilers)
                property_updated = True

            elif "PreinfusionSettings" in msg:
                settings: dict[str, Any] = {}
                settings["preinfusionSettings"] = json.loads(msg["PreinfusionSettings"])

                self.config.prebrew_mode, self.config.prebrew_configuration = (
                    parse_preinfusion_settings(settings)
                )
                property_updated = True

            elif "SystemInfo" in msg:
                self._system_info = json.loads(msg["SystemInfo"])
                property_updated = True

            elif "TankStatus" in msg:
                self.config.water_contact = msg["TankStatus"] == "full"  # or "empty"
                property_updated = True

        if not property_updated:
            raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")
        return True



"""Models for La Marzocco API"""

from dataclasses import dataclass
from typing import TypedDict

from .const import (
    BoilerType,
    PhysicalKey,
    PrebrewMode,
    SmartStandbyMode,
    SteamLevel,
    WeekDay,
)

####################################
#### base iot device specific ######
####################################


@dataclass(kw_only=True)
class LaMarzoccoFirmware:
    """Class for La Marzocco machine firmware"""

    current_version: str
    latest_version: str


@dataclass(kw_only=True)
class LaMarzoccoStatistics:
    """Class for La Marzocco machine statistics"""


@dataclass(kw_only=True)
class LaMarzoccoDeviceConfig:
    """Class for La Marzocco device configuration"""

    turned_on: bool
    doses: dict[PhysicalKey, float]


####################################
###### machine specific ############
####################################


@dataclass(kw_only=True)
class LaMarzoccoBoiler:
    """Class for La Marzocco boiler"""

    enabled: bool
    current_temperature: float
    target_temperature: float


@dataclass(kw_only=True)
class LaMarzoccoCoffeeStatistics(LaMarzoccoStatistics):
    """Class for La Marzocco coffee machine statistics"""

    drink_stats: dict[PhysicalKey, int]
    continous: int
    total_flushes: int

    @property
    def total_coffee(self) -> int:
        """Return the total amount of coffee brewed"""
        return sum(self.drink_stats.values())


@dataclass(kw_only=True)
class LaMarzoccoPrebrewConfiguration:
    """Class for La Marzocco key configuration"""

    on_time: float
    off_time: float

    @property
    def preinfusion_time(self) -> float:
        """Prefinfusion time is off time"""
        return self.off_time


@dataclass(kw_only=True)
class LaMarzoccoSmartStandby:
    """Class for La Marzocco smart standby configuration"""

    enabled: bool
    minutes: int
    mode: SmartStandbyMode


@dataclass(kw_only=True)
class LaMarzoccoWakeUpSleepEntry:
    """Class for La Marzocco wake up sleep entry"""

    enabled: bool
    days: list[WeekDay]
    entry_id: str
    steam: bool
    time_on: str
    time_off: str


# @dataclass(kw_only=True)
# class LaMarzoccoScheduleDay:
#     """Class for La Marzocco schedule day"""

#     enabled: bool
#     h_on: int
#     h_off: int
#     m_on: int
#     m_off: int


# @dataclass(kw_only=True)
# class LaMarzoccoSchedule:
#     """Class for La Marzocco schedule"""

#     enabled: bool
#     days: dict[WeekDay, LaMarzoccoScheduleDay]


@dataclass(kw_only=True)
class LaMarzoccoMachineConfig(LaMarzoccoDeviceConfig):
    """Class for La Marzocco machine configuration"""

    boilers: dict[BoilerType, LaMarzoccoBoiler]
    prebrew_mode: PrebrewMode
    plumbed_in: bool
    prebrew_configuration: dict[PhysicalKey, LaMarzoccoPrebrewConfiguration]
    dose_hot_water: int
    water_contact: bool
    wake_up_sleep_entries: dict[str, LaMarzoccoWakeUpSleepEntry]
    smart_standby: LaMarzoccoSmartStandby
    brew_active: bool
    brew_active_duration: float
    backflush_enabled: bool

    @property
    def steam_level(self) -> SteamLevel:
        """Return the steam level"""

        steam_boiler = self.boilers[BoilerType.STEAM]
        return min(SteamLevel, key=lambda x: abs(x - steam_boiler.target_temperature))


####################################
###### cloud client specific #######
####################################
@dataclass(kw_only=True)
class LaMarzoccoDeviceInfo:
    """Class for La Marzocco machine information."""

    serial_number: str
    name: str
    communication_key: str
    model: str


class LaMarzoccoCloudScheduleDay(TypedDict):
    """Input object to set the schedule"""

    day: str
    enable: bool
    on: str
    off: str


class LaMarzoccoCloudSchedule(TypedDict):
    """Input object to set the schedule"""

    enable: bool
    days: list[LaMarzoccoCloudScheduleDay]


@dataclass
class AccessToken:
    """Class for OAuth access token"""

    access_token: str
    expires_in: float
    refresh_token: str


####################################
###### grinder specific ############
####################################
@dataclass(kw_only=True)
class LaMarzoccoGrinderConfig(LaMarzoccoDeviceConfig):
    """Class for La Marzocco grinder configuration"""

    turned_on: bool
    led_enabled: bool
    bell_opened: bool
    stand_by_time: int



"""Base class for all La Marzocco IoT devices."""

import asyncio
import logging
from abc import abstractmethod
from typing import Any

from bleak import BleakError

from .client_bluetooth import LaMarzoccoBluetoothClient
from .client_cloud import LaMarzoccoCloudClient
from .client_local import LaMarzoccoLocalClient
from .const import FirmwareType
from .exceptions import (
    AuthFail,
    BluetoothConnectionFailed,
    ClientNotInitialized,
    RequestNotSuccessful,
)
from .helpers import parse_firmware
from .models import LaMarzoccoDeviceConfig, LaMarzoccoFirmware, LaMarzoccoStatistics

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoDevice:
    """Base class for all La Marzocco devices"""

    _cloud_client: LaMarzoccoCloudClient | None

    def __init__(
        self,
        model: str,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
        bluetooth_client: LaMarzoccoBluetoothClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoMachine instance"""

        self.model: str = model
        self.serial_number: str = serial_number
        self.name: str = name
        self.firmware: dict[FirmwareType, LaMarzoccoFirmware] = {}
        self.statistics: LaMarzoccoStatistics | None = None
        self.config: LaMarzoccoDeviceConfig = LaMarzoccoDeviceConfig(
            turned_on=False,
            doses={},
        )

        self._raw_config: dict[str, Any] | None = None
        self._cloud_client = cloud_client
        self._bluetooth_client = bluetooth_client
        self._local_client = local_client
        self._update_lock = asyncio.Lock()
        self._local_connection_down = False

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        self.firmware = parse_firmware(raw_config["firmwareVersions"], self.firmware)

    @property
    def cloud_client(self) -> LaMarzoccoCloudClient:
        """Ensure that the cloud client is initialized."""

        if self._cloud_client is None:
            raise ClientNotInitialized("Cloud client not initialized")
        return self._cloud_client

    @abstractmethod
    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

    @property
    def bluetooth_connected(self) -> bool:
        """Return the connection status of the bluetooth client."""

        if self._bluetooth_client is None:
            return False
        return self._bluetooth_client.connected

    @property
    def full_model_name(self) -> str:
        """Return the full model name of the device."""
        return self.model

    async def get_config(
        self,
        local_api_retry_delay: int = 3,
    ) -> None:
        """Update the machine status."""

        raw_config: dict[str, Any] = {}

        async with self._update_lock:
            # first, try to update locally
            if self._local_client is not None:
                try:
                    raw_config = await self._local_client.get_config()
                except AuthFail as exc:
                    _LOGGER.debug(
                        "Got 403 from local API, sending token request to cloud"
                    )
                    # for some machines (GS3) we need to send a token request before local API works
                    if self._cloud_client is not None:
                        await self._cloud_client.token_command(self.serial_number)
                        await asyncio.sleep(local_api_retry_delay)
                        raw_config = await self._local_client.get_config()
                    else:
                        raise exc
                except RequestNotSuccessful as exc:
                    msg = (
                        "Could not connect to local API although initialized, "
                        "falling back to cloud."
                    )
                    if not self._local_connection_down:
                        _LOGGER.warning(msg)
                        self._local_connection_down = True
                    _LOGGER.debug(msg)
                    _LOGGER.debug(exc)
                    if self._cloud_client is None:
                        raise exc
                else:
                    if self._local_connection_down:
                        _LOGGER.warning("Local API connection restored.")
                        self._local_connection_down = False

            # if local update failed, try to update from cloud
            if self._cloud_client is not None and not raw_config:
                raw_config = await self._cloud_client.get_config(self.serial_number)

            self.parse_config(raw_config)

    async def get_statistics(self) -> None:
        """Update the statistics"""

        raw_statistics = await self.cloud_client.get_statistics(self.serial_number)
        self.parse_statistics(raw_statistics)

    async def get_firmware(self) -> None:
        """Update the firmware"""

        self.firmware = await self.cloud_client.get_firmware(self.serial_number)

    @abstractmethod
    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""

    async def _bluetooth_command_with_cloud_fallback(
        self,
        command: str,
        **kwargs,
    ) -> bool:
        """Send a command to the machine via Bluetooth, falling back to cloud if necessary."""

        # First, try with bluetooth
        if self._bluetooth_client is not None:
            func = getattr(self._bluetooth_client, command)
            try:
                _LOGGER.debug(
                    "Sending command %s over bluetooth with params %s",
                    command,
                    str(kwargs),
                )
                await func(**kwargs)
            except (BleakError, BluetoothConnectionFailed) as exc:
                msg = "Could not send command to bluetooth device, even though initalized."

                if self._cloud_client is None:
                    _LOGGER.error(
                        "%s Cloud client not initialized, cannot fallback. Full error %s",
                        msg,
                        exc,
                    )
                    return False

                _LOGGER.warning("%s Falling back to cloud", msg)
                _LOGGER.debug("Full error: %s", exc)
            else:
                return True

        # no bluetooth or failed, try with cloud
        if self._cloud_client is not None:
            _LOGGER.debug(
                "Sending command %s over cloud with params %s",
                command,
                str(kwargs),
            )
            func = getattr(self._cloud_client, command)
            kwargs["serial_number"] = self.serial_number
            if await func(**kwargs):
                return True
        return False

    def __str__(self) -> str:
        config = dict(vars(self.config).items())
        return str(config)



"""Exceptions for the pylamarzocco package."""


class AuthFail(Exception):
    """Error to indicate there is invalid auth info."""


class RequestNotSuccessful(Exception):
    """Error to indicate didn't get a 200 response"""


class BluetoothConnectionFailed(Exception):
    """Error to indicate we couldn't successfully connect through bluetooth"""


class UnknownWebSocketMessage(Exception):
    """Error to indicate we received an unknown websocket message"""


class ClientNotInitialized(Exception):
    """Error to indicate the client is not initialized"""



"""Test the LaMarzoccoMachine class."""

# pylint: disable=W0212

from dataclasses import asdict
from http import HTTPMethod
from unittest.mock import AsyncMock, patch

from aioresponses import aioresponses
from syrupy import SnapshotAssertion

from pylamarzocco.client_bluetooth import LaMarzoccoBluetoothClient
from pylamarzocco.client_cloud import LaMarzoccoCloudClient
from pylamarzocco.client_local import LaMarzoccoLocalClient
from pylamarzocco.const import BoilerType, PhysicalKey

from . import init_machine
from .conftest import load_fixture


async def test_create(
    cloud_client: LaMarzoccoCloudClient,
    snapshot: SnapshotAssertion,
) -> None:
    """Test creation of a cloud client."""

    machine = await init_machine(cloud_client)
    assert asdict(machine.config) == snapshot(name="config")
    assert machine.firmware == snapshot(name="firmware")
    assert machine.statistics == snapshot(name="statistics")


async def test_local_client(
    local_machine_client: LaMarzoccoLocalClient,
    cloud_client: LaMarzoccoCloudClient,
    mock_aioresponse: aioresponses,
) -> None:
    """Ensure that the local client delivers same result"""
    # load config
    mock_aioresponse.get(
        url="http://192.168.1.42:8081/api/v1/config",
        status=200,
        payload=load_fixture("machine", "config.json")["data"],
    )

    machine = await init_machine(local_client=local_machine_client)

    machine2 = await init_machine(cloud_client)

    assert machine
    assert str(machine.config) == str(machine2.config)


async def test_set_temp(
    cloud_client: LaMarzoccoCloudClient,
) -> None:
    """Test setting boiler temperature."""
    machine = await init_machine(cloud_client)

    result = await machine.set_temp(
        BoilerType.STEAM,
        120,
    )
    assert result is True
    assert machine.config.boilers[BoilerType.STEAM].target_temperature == 120


# async def test_set_schedule(
#     cloud_client: LaMarzoccoCloudClient,
# ) -> None:
#     """Test setting prebrew infusion."""
#     machine = await init_machine(cloud_client)

#     with patch("asyncio.sleep", new_callable=AsyncMock):
#         result = await machine.set_schedule_day(
#             day=WeekDay.MONDAY,
#             enabled=True,
#             h_on=3,
#             m_on=0,
#             h_off=24,
#             m_off=0,
#         )
#     assert result is True


async def test_websocket_message(
    cloud_client: LaMarzoccoCloudClient,
    local_machine_client: LaMarzoccoLocalClient,
    snapshot: SnapshotAssertion,
):
    """Test parsing of websocket messages."""
    machine = await init_machine(cloud_client, local_client=local_machine_client)

    message = r'[{"Boilers":"[{\"id\":\"SteamBoiler\",\"isEnabled\":true,\"target\":131,\"current\":113},{\"id\":\"CoffeeBoiler1\",\"isEnabled\":true,\"target\":94,\"current\":81}]"}]'
    machine.on_websocket_message_received(message)
    assert asdict(machine.config) == snapshot

    message = r'[{"BoilersTargetTemperature":"{\"SteamBoiler\":131,\"CoffeeBoiler1\":94}"},{"Boilers":"[{\"id\":\"SteamBoiler\",\"isEnabled\":true,\"target\":131,\"current\":50},{\"id\":\"CoffeeBoiler1\",\"isEnabled\":true,\"target\":94,\"current\":36}]"}]'
    machine.on_websocket_message_received(message)
    assert asdict(machine.config) == snapshot


async def test_set_power(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting the power."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    assert await machine.set_power(True)

    bluetooth_client._client.write_gatt_char.assert_called_once_with(  # type: ignore[attr-defined]
        "050b7847-e12b-09a8-b04b-8e0922a9abab",
        b'{"name":"MachineChangeMode","parameter":{"mode":"BrewingMode"}}\x00',
    )
    mock_aioresponse.assert_any_call(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/status",
        json={"status": "BrewingMode"},
        timeout=5,
        headers={"Authorization": "Bearer token"},
    )


async def test_set_steam(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting the steam."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    assert await machine.set_steam(True)

    bluetooth_client._client.write_gatt_char.assert_called_once_with(  # type: ignore[attr-defined]
        "050b7847-e12b-09a8-b04b-8e0922a9abab",
        b'{"name":"SettingBoilerEnable","parameter":{"identifier":"SteamBoiler","state":true}}\x00',
    )
    mock_aioresponse.assert_any_call(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/enable-boiler",
        json={"identifier": "SteamBoiler", "state": True},
        timeout=5,
        headers={"Authorization": "Bearer token"},
    )


async def test_set_temperature(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting temperature."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    assert await machine.set_temp(BoilerType.STEAM, 131)

    bluetooth_client._client.write_gatt_char.assert_called_once_with(  # type: ignore[attr-defined]
        "050b7847-e12b-09a8-b04b-8e0922a9abab",
        b'{"name":"SettingBoilerTarget","parameter":{"identifier":"SteamBoiler","value":131}}\x00',
    )
    mock_aioresponse.assert_any_call(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/target-boiler",
        json={"identifier": "SteamBoiler", "value": 131},
        timeout=5,
        headers={"Authorization": "Bearer token"},
    )


async def test_set_prebrew_time(
    cloud_client: LaMarzoccoCloudClient,
    mock_aioresponse: aioresponses,
):
    """Test setting prebrew time."""
    machine = await init_machine(
        cloud_client,
    )

    assert await machine.set_prebrew_time(1.0, 3.5)

    mock_aioresponse.assert_any_call(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/setting-preinfusion",
        json={
            "button": "DoseA",
            "group": "Group1",
            "holdTimeMs": 3500,
            "wetTimeMs": 1000,
        },
        timeout=5,
        headers={"Authorization": "Bearer token"},
    )

    assert machine.config.prebrew_configuration[PhysicalKey.A].on_time == 1.0
    assert machine.config.prebrew_configuration[PhysicalKey.A].off_time == 3.5


async def test_set_preinfusion_time(
    cloud_client: LaMarzoccoCloudClient,
    mock_aioresponse: aioresponses,
):
    """Test setting prebrew time."""
    machine = await init_machine(
        cloud_client,
    )
    assert await machine.set_preinfusion_time(4.5)
    mock_aioresponse.assert_any_call(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/setting-preinfusion",
        json={"button": "DoseA", "group": "Group1", "holdTimeMs": 4500, "wetTimeMs": 0},
        timeout=5,
        headers={"Authorization": "Bearer token"},
    )

    assert machine.config.prebrew_configuration[PhysicalKey.A].off_time == 4.5



"""La Marzocco grinder module."""

from __future__ import annotations

from typing import Any

from .client_cloud import LaMarzoccoCloudClient
from .client_local import LaMarzoccoLocalClient
from .const import FirmwareType, GrinderModel, PhysicalKey
from .lm_device import LaMarzoccoDevice
from .models import LaMarzoccoGrinderConfig


class LaMarzoccoGrinder(LaMarzoccoDevice):
    """Class for La Marzocco grinder"""

    def __init__(
        self,
        model: GrinderModel,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoGrinder instance."""

        super().__init__(model, serial_number, name, cloud_client, local_client)
        self.config: LaMarzoccoGrinderConfig = LaMarzoccoGrinderConfig(
            turned_on=False,
            led_enabled=False,
            bell_opened=False,
            stand_by_time=0,
            doses={},
        )

    @classmethod
    async def create(cls, *args, **kwargs) -> LaMarzoccoGrinder:
        """Create a new LaMarzoccoGrinder instance."""

        self = LaMarzoccoGrinder(*args, **kwargs)
        await self.get_config()
        return self

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        super().parse_config(raw_config)
        self.config.turned_on = raw_config["machineMode"] == "StandBy"
        self.config.led_enabled = raw_config["baristaLed"]
        self.config.bell_opened = raw_config["bellOpened"]
        self.config.stand_by_time = raw_config["standByTime"]
        for dose in raw_config["doses"]:
            self.config.doses[PhysicalKey(ord(dose["doseIndex"][-1]) - 64)] = dose[
                "target"
            ]

    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

        raise NotImplementedError

    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""
        raise NotImplementedError
